AddonProfiles = ZO_CallbackObject:New()

local myNAME = "AddonProfiles"
local NO_PROFILE = ""
local NUM_ADDON_PROFILES_BINDINGS = 8
local STR_PROFILE_NONE = GetString(SI_ADDON_PROFILES_PROFILE_NONE)
local STR_SWITCH_TO_PROFILE = GetString(SI_ADDON_PROFILES_SWITCH_TO_PROFILE_FORMAT)
local ADDON_MANAGER = GetAddOnManager()
local g_savedVars =
{
    profiles = {},
    characterProfiles = {},
    keyboundProfiles = {},
}
local g_profileCombo = nil
local g_widget = nil


for i = 1, NUM_ADDON_PROFILES_BINDINGS do
    local name = "SI_BINDING_NAME_SWITCH_TO_PROFILE_" .. i
    ZO_CreateStringId(name, zo_strformat(STR_SWITCH_TO_PROFILE, i))
end


local function debug(fmt, ...)
    --d(string.format(fmt, ...))
end


local function warn(fmt, ...)
    d(string.format(fmt, ...))
end


local function applyProfileSettings(profileName)
    local profileSettings = g_savedVars.profiles[profileName]
    local dirty = false

    if type(profileSettings) == "table" then
        for i = 1, ADDON_MANAGER:GetNumAddOns() do
            local name, _, _, _, enabled = ADDON_MANAGER:GetAddOnInfo(i)
            local shouldBeEnabled = profileSettings[name] and true or false
            if shouldBeEnabled ~= enabled then
                ADDON_MANAGER:SetAddOnEnabled(i, shouldBeEnabled)
                dirty = true
            end
        end
    end

    return dirty
end


local function getCurrentProfile()
    local currentFilter = ADDON_MANAGER:GetAddOnFilter()
    local currentProfile = g_savedVars.characterProfiles[currentFilter]

    if not g_savedVars.profiles[currentProfile] then
        currentProfile = NO_PROFILE
        g_savedVars.characterProfiles[currentFilter] = nil
    end

    return currentProfile, currentFilter
end


local function refreshProfileSettings(profileName)
    local indexedSettings = {}
    local profileSettings = g_savedVars.profiles[profileName]

    if type(profileSettings) ~= "table" then
        profileSettings = {}
        g_savedVars.profiles[profileName] = profileSettings
    end

    for i = 1, ADDON_MANAGER:GetNumAddOns() do
        local name, _, _, _, enabled = ADDON_MANAGER:GetAddOnInfo(i)
        indexedSettings[i] = enabled
        profileSettings[name] = enabled
    end

    return indexedSettings
end


local function setAddOnEnabledForAptCharacters(data, enabled)
    local currentFilter = ADDON_MANAGER:GetAddOnFilter()
    local currentProfile = g_savedVars.characterProfiles[currentFilter]
    local profileSettings = g_savedVars.profiles[currentProfile]

    if type(profileSettings) == "table" then
        profileSettings[data.addOnFileName] = enabled
    else
        return
    end

    for filter, profile in pairs(g_savedVars.characterProfiles) do
        if profile == currentProfile then
            debug("setting [%d] %s = %s for %s",
                  data.index,
                  data.addOnFileName,
                  enabled and "enabled" or "disabled",
                  tostring(filter))
            ADDON_MANAGER:SetAddOnFilter(filter)
            ADDON_MANAGER:SetAddOnEnabled(data.index, enabled)
        end
    end

    debug("restoring filter: %s", tostring(currentFilter))
    ADDON_MANAGER:SetAddOnFilter(currentFilter)
end


local function setCharacterProfile(characterFilter, profileName)
    if not characterFilter then
        characterFilter = ADDON_MANAGER:GetAddOnFilter()
    end
    g_savedVars.characterProfiles[characterFilter] = profileName
    AddonProfiles:FireCallbacks("ProfileChanged", profileName)
end


local function onSelectProfile(profileCombo, _, item, selectionChanged)
    local setDirty = false
    local profileName = item.profileName

    debug("select profile: %s (%s)", profileName, tostring(selectionChanged))

    if selectionChanged then
        setCharacterProfile(nil, profileName)
        setDirty = applyProfileSettings(profileName)
    end

    if ADD_ON_MANAGER and setDirty then
        ADD_ON_MANAGER.isDirty = true
        ADD_ON_MANAGER:RefreshData()
        ADD_ON_MANAGER:RefreshMultiButton()
    end
end


local function initControls()
    g_widget = CreateControlFromVirtual("$(parent)_AddonProfiles", ZO_AddOns,
                                        "AddonProfilesTemplate")
    g_widget:SetAnchor(LEFT, ZO_AddOnsTitle, RIGHT, 20, 0)

    local function createNewProfile()
		ZO_Dialogs_ShowDialog("AddonProfiles_CreateProfile")
    end

    local function deleteSelectedProfile()
        local item = g_profileCombo:GetSelectedItemData()
        if item then
            ZO_Dialogs_ShowDialog("AddonProfiles_DeleteProfile", item, {
                mainTextParams = { item.highlightName }
            })
        end
    end

    local function renameSelectedProfile()
        -- TODO
    end

    local function onOptionsButtonClicked(control, button)
        if button == 1 then -- left mouse button
            ClearMenu()
            AddCustomMenuItem(GetString(SI_ADDON_PROFILES_MENU_CREATE_PROFILE), createNewProfile)
            local item = g_profileCombo:GetSelectedItemData()
            if item and item.sortGroup == 2 then
                AddCustomMenuItem(GetString(SI_ADDON_PROFILES_MENU_DELETE_PROFILE), deleteSelectedProfile)
            end
            ShowMenu(control)
            AnchorMenu(control, 1)
        end
    end

    g_widget.optionsButton = g_widget:GetNamedChild("_OptionsButton")
    g_widget.optionsButton:SetHandler("OnClicked", onOptionsButtonClicked)

    local currentProfile = getCurrentProfile()
    local dropdownControl = g_widget:GetNamedChild("_SetProfileDropdown")
    g_profileCombo = ZO_ComboBox_ObjectFromContainer(dropdownControl)
    g_profileCombo:AddItem2(1, NO_PROFILE, STR_PROFILE_NONE, onSelectProfile, true)
    g_savedVars.profiles[NO_PROFILE] = nil -- should never appear among user profiles

    for profileName in pairs(g_savedVars.profiles) do
        local selected = (profileName == currentProfile)
        g_profileCombo:AddItem2(2, profileName, profileName, onSelectProfile, selected)
    end

    g_profileCombo:UpdateItems()
end


local function initSavedVars()
    if type(AddonProfiles_SavedVariables2) ~= "table" then
        AddonProfiles_SavedVariables2 = g_savedVars
    else
        for k, v in pairs(g_savedVars) do
            if type(AddonProfiles_SavedVariables2[k]) ~= type(v) then
                AddonProfiles_SavedVariables2[k] = v
            end
        end
        g_savedVars = AddonProfiles_SavedVariables2
    end
end


local function hookAddOnManager()
    local zorgAddOnManager_GetRowSetupFunction = ADD_ON_MANAGER.GetRowSetupFunction
    local zorgAddOnManager_OnEnabledButtonClicked = ADD_ON_MANAGER.OnEnabledButtonClicked

    local function wrapRowSetupFunction(orgSetup)
        return function(rowControl, data, list)
            orgSetup(rowControl, data, list)

            local currentProfile = getCurrentProfile()
            local profileSettings = g_savedVars.profiles[currentProfile]
            local mismatchControl = rowControl:GetNamedChild("_AddonProfilesMismatch")
            local mismatch = false

            if profileSettings then
                local disabledInProfile = not profileSettings[data.addOnFileName]
                mismatch = disabledInProfile == data.addOnEnabled
            end

            if not mismatch then
                if mismatchControl then
                    mismatchControl:SetHidden(true)
                end
            elseif mismatchControl then
                mismatchControl:SetHidden(false)
            else
                local enabledControl = rowControl:GetNamedChild("Enabled")
                mismatchControl = rowControl:CreateControl("$(parent)_AddonProfilesMismatch",
                                                           CT_TEXTURE)
                mismatchControl:SetAnchor(TOPLEFT, enabledControl, TOPLEFT, -5, -3)
                mismatchControl:SetAnchor(BOTTOMRIGHT, enabledControl, BOTTOMRIGHT, 3, 4)
                mismatchControl:SetColor(ZO_ERROR_COLOR:UnpackRGBA())
                mismatchControl:SetDrawLayer(DL_BACKGROUND)
                mismatchControl:SetDrawLevel(10)
            end
        end
    end

    function ADD_ON_MANAGER:GetRowSetupFunction()
        local orgSetup = zorgAddOnManager_GetRowSetupFunction(self)
        return wrapRowSetupFunction(orgSetup)
    end

    for typeId, dataType in next, ADD_ON_MANAGER.list.dataTypes do
        dataType.setupCallback = wrapRowSetupFunction(dataType.setupCallback)
    end

    function ADD_ON_MANAGER:OnEnabledButtonClicked(control, checkState)
        if checkState ~= TRISTATE_CHECK_BUTTON_INDETERMINATE then
            local rowControl = control:GetParent()
            local data = ZO_ScrollList_GetData(rowControl)
            local enabled = (checkState == TRISTATE_CHECK_BUTTON_CHECKED)
            setAddOnEnabledForAptCharacters(data, enabled)
        end
        -- original handler rebuilds the list, possibly assigning another
        -- add-on to the clicked row, so it must called last
        zorgAddOnManager_OnEnabledButtonClicked(self, control, checkState)
    end
end


local function hookKeybindingManager()

    local function assignProfileToKeybind(comboBox, _, newItem, selectionChanged)
        local actionName = comboBox.keybindData.actionName
        local oldProfile = g_savedVars.keyboundProfiles[actionName]
        local newProfile = newItem.profileName

        if newProfile == NO_PROFILE then
            g_savedVars.keyboundProfiles[actionName] = nil
        else
            g_savedVars.keyboundProfiles[actionName] = newProfile
            -- if the selected profile is already assigned to another keybind,
            -- swap it for the one being replaced
            for action, profile in pairs(g_savedVars.keyboundProfiles) do
                if action ~= actionName and profile == newProfile then
                    g_savedVars.keyboundProfiles[action] = oldProfile
                    ZO_ScrollList_RefreshVisible(ZO_KeybindingsList)
                    break
                end
            end
        end
    end

    local function getProfileComboBox(rowControl, actionName)
        local index = actionName:match("^SWITCH_TO_PROFILE_(%d+)$")
        if index then
            local name = "_AddonProfilesSwitch" .. index
            local parent = rowControl:GetParent()
            local control = parent:GetNamedChild(name)
            if not control then
                control = CreateControlFromVirtual("$(parent)" .. name, parent,
                                                   "AddonProfilesComboBox")
            end
            return control
        end
    end

    local function setupRow(rowControl, data)
        local profileControl = getProfileComboBox(rowControl, data.actionName)
        if not profileControl then
            return
        end

        rowControl.actionLabel:SetHidden(true)
        profileControl:SetHidden(false)
        profileControl:SetAnchor(TOPLEFT, rowControl, TOPLEFT, 5, 0)

        local primaryBind = rowControl.bindingButtons[1]
        if VOTANS_KEYBINDER then
            profileControl:SetAnchor(BOTTOMRIGHT, primaryBind, BOTTOMLEFT, -30, 0)
        else
            profileControl:SetAnchor(BOTTOMRIGHT, primaryBind, BOTTOMLEFT, -5, 0)
        end

        local assignedProfile = g_savedVars.keyboundProfiles[data.actionName]
        local profileCombo = ZO_ComboBox_ObjectFromContainer(profileControl)
        local selected = true

        profileCombo.keybindData = data
        profileCombo:ClearItems()

        local itemName = zo_strformat(STR_SWITCH_TO_PROFILE, STR_PROFILE_NONE)
        profileCombo:AddItem2(1, NO_PROFILE, itemName, assignProfileToKeybind, selected)

        for profileName in pairs(g_savedVars.profiles) do
            selected = (profileName == assignedProfile)
            itemName = zo_strformat(STR_SWITCH_TO_PROFILE, profileName)
            profileCombo:AddItem2(2, profileName, itemName, assignProfileToKeybind, selected)
        end

        profileCombo:UpdateItems()
    end

    local function hideRow(rowControl, data)
        local profileControl = getProfileComboBox(rowControl, data.actionName)
        if not profileControl then
            return
        end

        rowControl.actionLabel:SetHidden(false)
        profileControl:SetHidden(true)
        profileControl:ClearAnchors()
    end

    CALLBACK_MANAGER:RegisterCallback("libAddonKeybinds.SetupKeybindRow", setupRow)
    CALLBACK_MANAGER:RegisterCallback("libAddonKeybinds.HideKeybindRow", hideRow)
end


local function onLoad(event, name)
    if name ~= myNAME then return end
    EVENT_MANAGER:UnregisterForEvent(myNAME, event)

    initSavedVars()
    initControls()

    hookAddOnManager()
    hookKeybindingManager()
end


EVENT_MANAGER:RegisterForEvent(myNAME, EVENT_ADD_ON_LOADED, onLoad)


---------------------
-- EXPOSED INTERFACE


function AddonProfiles:CreateProfile(profileName)
    if g_savedVars.profiles[profileName] or profileName == NO_PROFILE then
        warn("profile already exists: %s", profileName)
    else
        refreshProfileSettings(profileName)
        self:FireCallbacks("ProfileCreated", profileName)
        g_profileCombo:AddItem2(2, profileName, profileName, onSelectProfile, true)
        g_profileCombo:UpdateItems()
        setCharacterProfile(nil, profileName)
        ZO_ScrollList_RefreshVisible(ZO_KeybindingsList)
    end
end


function AddonProfiles:DeleteProfile(profileName)
    if profileName ~= NO_PROFILE then
        g_savedVars.profiles[profileName] = nil
        local item = g_profileCombo:RemoveItemByKey("profileName", profileName)
        if item == g_profileCombo:GetSelectedItemData() then
            g_profileCombo:SelectFirstItem()
        end
        self:FireCallbacks("ProfileDeleted", profileName)
        ZO_ScrollList_RefreshVisible(ZO_KeybindingsList)
    end
end


function AddonProfiles:GetCurrentProfile()
    return getCurrentProfile()
end


function AddonProfiles:SwitchToProfile(keybind)
    local currentProfile, currentFilter = getCurrentProfile()
    local profileName = g_savedVars.keyboundProfiles[keybind]
    if not g_savedVars.profiles[profileName] then
        profileName = NO_PROFILE
    end
    if profileName ~= currentProfile then
        g_profileCombo:SelectItemByKey("profileName", profileName, true)
        setCharacterProfile(currentFilter, profileName)
    end
    if applyProfileSettings(profileName) then
        ReloadUI()
    end
end
